#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>
from .loss import Loss
from .callback import Callback
from .optim import Optimizer
from .scheduler import Scheduler
from .checkpoint import CheckpointManager
from .trainer import Trainer



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>
import torch
from torch import nn, Tensor
from torch.nn.utils.rnn import pack_padded_sequence, pad_packed_sequence


class RNN(nn.Module):
    __RNN__ = nn.RNN

    def __init__(self, input_size, hidden_size, num_layers=1,
                 bias=True, batch_first=True, dropout=0., bidirectional=True, **kwargs):
        super(RNN, self).__init__()
        self.rnn = self.__RNN__(
            input_size=input_size,
            hidden_size=hidden_size,
            num_layers=num_layers,
            bias=bias,
            batch_first=batch_first,
            dropout=dropout,
            bidirectional=bidirectional,
            **kwargs
        )

    @property
    def batch_first(self):
        return self.rnn.batch_first

    @property
    def batch_dim(self):
        return 0 if self.batch_first else 1

    @property
    def seq_dim(self):
        return 1 if self.batch_first else 0

    @property
    def num_layers(self):
        return self.rnn.num_layers

    @property
    def bidirectional(self):
        return self.rnn.bidirectional

    @property
    def hidden_size(self):
        return self.rnn.hidden_size

    @property
    def output_size(self):
        return self.hidden_size * 2 if self.bidirectional else self.hidden_size

    @property
    def num_directions(self):
        return 2 if self.bidirectional else 1

    def forward(self, inputs: Tensor, hx=None, seq_lens: Tensor = None, process=False):
        if seq_lens is None:
            batch_num, seq_length = inputs.shape[self.batch_dim], inputs.shape[self.seq_dim]
            seq_lens = torch.full((batch_num,), seq_length, device=inputs.device)

        # sort
        sort_lens, sort_idx = torch.sort(seq_lens, dim=0, descending=True)
        inputs = torch.index_select(inputs, dim=self.batch_dim, index=sort_idx)

        # pack
        packed_inputs = pack_padded_sequence(inputs, sort_lens, batch_first=self.batch_first)
        packed_outputs, hidden_last = self.rnn(packed_inputs, hx=hx)  # -> [N,L,C]

        # unpack
        outputs, lengths = pad_packed_sequence(packed_outputs, batch_first=self.batch_first)

        # unsort
        _, unsort_idx = torch.sort(sort_idx, dim=0, descending=False)

        def unsort(inputs, dim=self.batch_dim):
            return torch.index_select(inputs, dim=dim, index=unsort_idx)

        if process:
            def process_fn(hidden):
                hidden = hidden.contiguous().view(
                    self.num_layers, self.num_directions,
                    inputs.shape[self.batch_dim], self.hidden_size)
                hidden = hidden[-1]  # (direction, batch, hidden/2) choose last layer
                if self.bidirectional:  # concat direction dim
                    return torch.cat((hidden[0], hidden[1]), -1)
                else:
                    return hidden[0]
        else:
            process_fn = lambda x: x

        if not isinstance(hidden_last, Tensor):
            hidden_last, cell_last = hidden_last
            return unsort(outputs), (unsort(process_fn(hidden_last), 1),
                                     unsort(process_fn(cell_last), 1))
        else:
            return unsort(outputs), unsort(process_fn(hidden_last), 1)


class LSTM(RNN):
    __RNN__ = nn.LSTM


class GRU(RNN):
    __RNN__ = nn.GRU



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>
import sys
from time import time
from typing import Iterable, Union, List, Set, Dict

import numpy as np
import torch
import torch.utils.data

from tqdm import tqdm
from ltp.exe.config import Config
from ltp.train.callback import Callback, ValidationCallback
from ltp.train.stop_condition import NoStopping, EarlyStopping
from ltp.eval.metrics import Metric
from ltp.train import Trainer
from ltp.utils import cycle


class Executor(object):
    """
    实际训练器的一个简单包装
    """
    config: Config
    trainer: Trainer
    __callbacks: List
    progressbar_metrics: Set
    stop_condition: Union[EarlyStopping, NoStopping]

    def __init__(self, config: Config):
        self.config = config
        self.stop_condition = EarlyStopping(
            **self.config.early_stopping
        ) if self.config.early_stopping else NoStopping()
        self.__callbacks = []
        self.progressbar_metrics = set()
        self.trainer = Trainer.from_params(self.config.trainer, config=self.config)
        self.tasks = self.config.tasks
        self.epoch_size = self.config.epoch_size

    def train(self, epoch: int = 30):
        epoch = self.config.epoch if epoch is None else epoch

        # =================================== 任务初始化 =================================
        multi_task_dataset = {}
        for name, task in self.tasks.items():
            if not task.config.dataset:
                continue
            if name != 'default':
                task.load()
            train, valid = task.train_dataset(self.config.batch_size)
            valid_callback = ValidationCallback(data_loader=valid, metrics=task.metrics, task=name)
            self.register_callback(valid_callback)
            multi_task_dataset[name] = train

        # =================================== 回调建立 ===================================
        for callback in self.config.callbacks:
            self.register_callback(Callback.from_params(callback))
        # ======================= 开始训练 ===============================================
        self.train_wrapper(multi_task_dataset, epoch, tau=self.config.tau)
        # =================================== 指标评估 ===================================
        self.evaluate()

    def evaluate(self, file: str = None, checkpoint: str = None, task: str = None):
        print("========================= Evaluate ==========================", file=sys.stderr)
        self.tasks['default'].load(self.trainer.state, checkpoint)
        if task is None:
            for name, task_obj in self.tasks.items():
                if not task_obj.config.dataset:
                    continue
                if name != 'default':
                    task_obj.load()
                test = task_obj.load_dataset(file, self.config.batch_size)
                self.evaluate_(test, task_obj.metrics, name)
        else:
            if task != 'default':
                self.tasks[task].load()
            test = self.tasks[task].load_dataset(file, self.config.batch_size)
            self.evaluate_(test, self.tasks[task].metrics, task)

    def predict(self, inputs: str, outputs: str, checkpoint: str = None, task: str = None):
        self.tasks['default'].load(self.trainer.state, checkpoint)
        if task != 'default':
            self.tasks[task].load()
        test = self.tasks[task].load_dataset(inputs, self.config.batch_size)
        self.predict_(test, outputs, task)

    def deploy(self, path: str = 'deploy.model', vocab: str = None):
        deploy_state_dict = {'vocab': vocab}
        for task, task_obj in self.tasks.items():
            task_obj.load()
            if task == 'default':
                deploy_state_dict['model'] = task_obj.model.state_dict()
                deploy_state_dict['model_config'] = task_obj.config.model
                deploy_state_dict['pretrained_config'] = task_obj.model.pretrained.config.to_dict()

            for field_name, field in task_obj.fields:
                if field.is_target and hasattr(field, 'vocab') and hasattr(field.vocab, 'itos'):
                    pad_bias = getattr(field, 'pad_bias', 0)
                    deploy_state_dict[task] = field.vocab.itos[pad_bias:]
        torch.save(deploy_state_dict, path)

    def train_wrapper(self, dataloaders: Dict[str, torch.utils.data.DataLoader], epochs: int = 30, tau: float = 1.0):
        """
        通过给定的 data loader 进行训练，训练会进行到epochs或者 stop condition = True

        :param tau: 放大指数
        :param dataloaders: PyTorch DataLoader
        :param epochs: 训练的最大轮数
        """
        print("========================== Train ============================", file=sys.stderr)
        self.trainer.model.train()  # set the module to training mode
        train_start = time()

        # cycle has a memory leak
        dataiters = {k: cycle(v) for k, v in dataloaders.items()}
        if all(hasattr(v, '__len__') for v in dataloaders.values()):
            dataloader_sizes = {k: len(v) for k, v in dataloaders.items()}
            total_size = sum(v for k, v in dataloader_sizes.items())
            Z = sum(pow(v, tau) for v in dataloader_sizes.values())
            tasknames, sampling_weights = zip(*((k, pow(v, tau) / Z) for k, v in dataloader_sizes.items()))
        else:
            raise NotImplementedError("Dataloader 需要实现 __len__ 方法")

        if self.epoch_size:
            total_size = self.epoch_size

        self.tasks['default'].build_scheduler(epochs * total_size)
        self.tasks['default'].restore(self.trainer.state)

        self.trainer.init(epochs * total_size)
        self.trainer.state.current_epoch += 1

        while self.trainer.state.current_epoch <= epochs and not self.stop_condition(self.trainer.state):
            # ------------------------- EPOCH ----------------------------
            self.trainer.before_train()
            self.train_(total_size, tasknames, sampling_weights, dataiters)
            self.__run_post_epoch_callbacks()
            self.trainer.after_train()
            # ------------------------- EPOCH ----------------------------
            self.trainer.state.current_epoch += 1
        print("train time %.2f" % (time() - train_start))

    def train_(self, total_size, tasknames, sampling_weights, dataiters):
        ofm = 0  # out of memory
        with tqdm(range(total_size), desc=f'Train({self.trainer.state.current_epoch}): ') as epoch_steps:
            for _ in epoch_steps:
                try:
                    taskname = np.random.choice(tasknames, p=sampling_weights)
                    dataiter = dataiters[taskname]
                    batch = next(dataiter)
                    self.trainer.state.last_train_loss = self.trainer.train(batch, task=taskname)
                    self.trainer.state.global_step += 1
                    self.__run_post_iteration_callbacks()

                    postfix = {metric: getattr(self.trainer.state, metric) for metric in
                               self.progressbar_metrics}
                    postfix["loss"] = self.trainer.state.last_train_loss
                    postfix["ofm"] = ofm
                    epoch_steps.set_postfix(postfix)
                except Exception as e:
                    detail = e.args[0]
                    if isinstance(detail, str) and detail.startswith("CUDA out of memory"):
                        ofm += 1
                        epoch_steps.set_postfix({"ofm": ofm})
                        continue
                    raise e

    def evaluate_(self, data_loader: torch.utils.data.DataLoader, metrics: Iterable[Metric], task: str) -> \
            Iterable[Metric]:
        """
        评估一个模型
        :param task: 任务信息
        :param data_loader:  PyTorch DataLoader
        :param metrics:  进行评估的 Metrics
        :return: 计算得到的各项指标
        """
        for metric in metrics:
            metric.clear()

        ofm = 0  # out of memory
        self.trainer.before_eval(task)
        with torch.no_grad(), tqdm(data_loader, desc=f'{task}({self.trainer.state.current_epoch}): ')as pbar:
            for batch in pbar:
                try:
                    x, y_pred, y = self.trainer.eval(batch, task)
                    metric_values = {}
                    for metric in metrics:
                        metric.step(y_pred, y)
                        metric_values.update(metric.compute())
                    if ofm > 0:
                        metric_values['ofm'] = ofm
                    pbar.set_postfix(metric_values)
                except Exception as e:
                    detail = e.args[0]
                    if isinstance(detail, str) and detail.startswith("CUDA out of memory"):
                        ofm += 1
                        continue
                    raise e
        self.trainer.after_eval(task)
        return metrics

    def predict_(self, dataloader: torch.utils.data.DataLoader, outputs: str, task):
        """
        进行预测操作

        :param dataloader: PyTorch DataLoader
        :param outputs: 输出的文件名
        """

        self.trainer.before_predict(task)
        with torch.no_grad(), open(outputs, mode='w', encoding='utf8') as f, \
                tqdm(dataloader, dynamic_ncols=True, desc=f'Predict: ') as pbar:
            for batch in pbar:
                result = self.trainer.predict(batch, task)
                for pred in result:
                    f.writelines("\t".join(pred) + "\n")
        self.trainer.after_predict(task)

    def test(self):
        for name, task in self.tasks.items():
            if not task.config.dataset:
                continue
            if name != 'default':
                task.load()
            train, valid = task.train_dataset(self.config.batch_size, False)

            for name, field in task.fields:
                if hasattr(field, 'vocab') and hasattr(field.vocab, 'itos'):
                    print(name, 'vocab size(with pad):', len(field.vocab.itos), file=sys.stderr)
                    print(field.vocab.itos, file=sys.stderr)

            for train_iter in tqdm(train):
                pass
            for valid_iter in tqdm(valid):
                pass
            for test_iter in tqdm(task.load_dataset(batch_size=self.config.batch_size)):
                pass

    def add_progressbar_metric(self, name):
        self.progressbar_metrics.add(name)

    def register_callback(self, callback: Callback):
        callback.init(self)
        self.__callbacks.append(callback)

    def __run_post_iteration_callbacks(self):
        for callback in self.__callbacks:
            if callback.iteration is None:
                continue
            if callback.iteration != 0 and self.trainer.state.global_step % callback.iteration == 0:
                callback(self)

    def __run_post_epoch_callbacks(self):
        for callback in self.__callbacks:
            if callback.epoch is None:
                continue
            if (callback.epoch != 0 and self.trainer.state.global_step % callback.epoch == 0):
                callback(self)



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>
import torch
import torch.nn
import torch.nn.functional as F
from torch.nn.modules.activation import *


class Swish(torch.nn.Module):
    r"""Swish activation function:

    .. math::
        \text{Swish}(x) = x * Sigmoid(x)

    Shape:
        - Input: :math:`(N, *)` where `*` means, any number of additional dimensions
        - Output: :math:`(N, *)`, same shape as the input
    """

    def forward(self, input):
        return input * torch.sigmoid(input)


class HSwish(torch.nn.Module):
    r"""Hard Swish activation function:

    .. math::
        \text{Swish}(x) = x * \frac{ReLU6(x+3)}{6}

    Shape:
        - Input: :math:`(N, *)` where `*` means, any number of additional dimensions
        - Output: :math:`(N, *)`, same shape as the input
    """

    def __init__(self):
        super().__init__()
        self.relu6 = ReLU6()

    def forward(self, input):
        return input * (self.relu6(input + 3) / 6)


class Mish(torch.nn.Module):
    r"""Mish activation function:

    .. math::
        \text{Mish}(x) = x * tanh(\ln(1 + e^x))

    Shape:
        - Input: :math:`(N, *)` where `*` means, any number of additional dimensions
        - Output: :math:`(N, *)`, same shape as the input
    """

    def forward(self, input):
        return input * (torch.tanh(F.softplus(input)))



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>
import sys
import os, regex, logging
from typing import List, Dict
from collections import OrderedDict

import torch

from ltp.core import Registrable
from ltp.const import checkpoint

default_save_diretory = './checkpoint'
default_best_filename = 'best.pt.tar'

logger = logging.getLogger(__name__)


class CheckpointPlugin(metaclass=Registrable):
    """
    Checkpoint

    :param directory: 存放目录
    :param filename: 检查点文件名
    """

    def __init__(self, directory: str = None, filename: str = None):

        if directory is None:
            directory = default_save_diretory
        if filename is None:
            filename = checkpoint

        self.directory = directory
        self.filename = filename
        os.makedirs(self.directory, exist_ok=True)

    def save(self, task, state):
        raise NotImplementedError

    def load(self, task, state):
        raise NotImplementedError

    def save_checkpoint(self, task, state, filename: str):
        filepath = os.path.join(self.directory, filename)
        ckpt_dict = {
            'model_state_dict': task.model.state_dict(),
            'optimizer_state_dict': task.optimizer.state_dict(),
            'executor_state': state,
            'check_model_class': str(task.model.__class__),
            'check_optimizer_class': str(task.optimizer.__class__),
            # 'check_ldtp_version': __version__    # TODO
        }

        if task.scheduler:
            ckpt_dict['scheduler_state_dict'] = task.scheduler.state_dict()
            ckpt_dict['check_scheduler_class'] = str(task.scheduler.__class__)

        for name, field in task.fields:
            if not hasattr(field, 'use_vocab') or not field.use_vocab:
                continue
            if hasattr(field, 'vocab') and hasattr(field.vocab, '__getstate__'):
                ckpt_dict[name] = field.vocab.__getstate__()

        torch.save(ckpt_dict, filepath)
        logger.debug(f"Have saved checkpoint to {filepath}!!!")
        return filepath

    def load_checkpoint(self, task, state, filename):
        filepath = os.path.join(self.directory, filename)

        model = task.model
        optimizer = task.optimizer

        if not os.path.isfile(filepath):
            logger.warning(f"Restored From checkpoint {filepath} failed!!!")
            return

        checkpoint = torch.load(filepath, map_location=task.device)
        model.load_state_dict(checkpoint['model_state_dict'])
        optimizer.load_state_dict(checkpoint['optimizer_state_dict'])
        if state:
            state.load_state_dict(checkpoint['executor_state'])

        if task.scheduler:
            task.scheduler.load_state_dict(checkpoint['scheduler_state_dict'])

        for name, field in task.fields:
            if not hasattr(field, 'use_vocab') or not field.use_vocab:
                continue
            field.build_vocab()
            if hasattr(field, 'vocab') and hasattr(field.vocab, '__setstate__'):
                field.vocab.__setstate__(checkpoint[name])

        logger.info(f"Have restored from checkpoint({filepath})")


class CheckpointManager(object):
    plugins: Dict[str, CheckpointPlugin]

    def __init__(self, directory=None, filename=None, plugins: List = None, **kwargs):
        """
        Checkpoint 管理器

        :param directory: Ckpt 保存路径
        :param filename: 保存的文件名，通常会被转成不同的形式
        :param plugins: List [best, restore]
        :param best: dict 传给 best 的额外参数，使用 best 插件时，metric 必选
        """
        if directory is None:
            directory = default_save_diretory
        self.directory = directory
        self.filename = filename
        self.plugins = OrderedDict()

        if plugins is not None:
            for plugin_name in plugins:
                if not hasattr(self, 'mode'):
                    self.mode = plugin_name
                plugin_class = CheckpointPlugin.by_name(plugin_name)
                plugin_config = kwargs.get(plugin_name, {})
                plugin_config.setdefault('filename', filename)
                plugin_config.setdefault('directory', directory)
                self.plugins[plugin_name] = plugin_class(**plugin_config)
        elif kwargs is None or not len(kwargs):
            self.mode = 'restore'
            self.plugins['restore'] = RestoreCheckpointPlugin(directory=directory, filename=filename)
        else:
            for plugin_name, plugin_config in kwargs.items():
                if not hasattr(self, 'mode'):
                    self.mode = plugin_name
                plugin_class = CheckpointPlugin.by_name(plugin_name)
                plugin_config.setdefault('filename', filename)
                plugin_config.setdefault('directory', directory)
                self.plugins[plugin_name] = plugin_class(**plugin_config)

    def __getitem__(self, item):
        if item in self.plugins:
            self.mode = item
        else:
            self.mode = LoadCheckpointPlugin(directory=self.directory, filename=item)
        return self

    def load(self, task, state):
        # 加载模型
        if isinstance(self.mode, str):
            self.plugins[self.mode].load(task, state)
        elif isinstance(self.mode, LoadCheckpointPlugin):
            self.mode.load(task, state)

    def save(self, task, state):
        for plugin in self.plugins.values():
            plugin.save(task, state)


class LoadCheckpointPlugin(CheckpointPlugin, alias='_load'):
    def load(self, task, state):
        self.load_checkpoint(task, state, self.filename)

    def save(self, task, state):
        pass


class BestCheckpointPlugin(CheckpointPlugin, alias='best'):
    def __init__(self, metric, directory: str = None, filename: str = None):
        if filename is None:
            filename = default_best_filename
        super(BestCheckpointPlugin, self).__init__(directory=directory, filename=filename)
        self.metric_name = metric
        self.best_metric = None

        c = self.filename.count('.')
        base, *ext = self.filename.rsplit('.', c)
        self.format = f"{base}_%.4f." + '.'.join(ext)
        # base_{metric}.ext
        self.regex = regex.compile(f"{base}_(\\d+\\.\\d+)\\." + '\\.'.join(ext))
        self.last_filepath = None

    def load(self, task, state):
        if not os.path.isdir(self.directory):
            return
        filenames = os.listdir(self.directory)

        best_metric = None
        best_checkpoint = None
        for filename in filenames:
            if os.path.isdir(filename):
                continue
            match = regex.fullmatch(self.regex, filename)
            if match is None:
                continue
            found_metric = float(match.group(1))

            if best_metric is None or found_metric > best_metric:
                best_metric = found_metric
                best_checkpoint = filename

        if best_checkpoint is not None:
            self.load_checkpoint(task, state, best_checkpoint)

    def save(self, task, state):
        try:
            if isinstance(self.metric_name, list):
                metric = sum([state.get(metric_name) for metric_name in self.metric_name]) / len(self.metric_name)
            else:
                metric = state.get(self.metric_name)
            if self.best_metric is None or metric > self.best_metric:
                self.best_metric = metric
                filename = self.format % self.best_metric
                if self.last_filepath is not None and os.path.exists(self.last_filepath):
                    os.remove(self.last_filepath)
                self.last_filepath = self.save_checkpoint(task, state, filename)
        except Exception as e:
            print(f"No metric {self.metric_name}")


class RestoreCheckpointPlugin(CheckpointPlugin, alias="restore"):
    def __init__(self, *args, **kwargs):
        super(RestoreCheckpointPlugin, self).__init__(*args, **kwargs)
        c = self.filename.count('.')
        base, *ext = self.filename.rsplit('.', c)
        self.format = f"{base}_%d." + '.'.join(ext)
        self.regex = regex.compile(f"{base}_(\\d+)\\." + '\\.'.join(ext))

    def load(self, task, state):
        last_iter = None
        last_checkpoint = None
        if not os.path.isdir(self.directory):
            return
        filenames = os.listdir(self.directory)
        for filename in filenames:
            if os.path.isdir(filename):
                continue
            match = regex.fullmatch(self.regex, filename)
            if match is None:
                continue

            found_epoch = int(match.group(1))
            if last_iter is None or found_epoch > last_iter:
                last_iter = found_epoch
                last_checkpoint = filename

        if last_checkpoint is not None:
            self.load_checkpoint(task, state, last_checkpoint)

    def save(self, task, state):
        filename = self.format % state.global_step
        self.save_checkpoint(task, state, filename)



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>

"""
IOBES 标签列表如下：

B，即Begin，表示开始
I，即Intermediate，表示中间
E，即End，表示结尾
S，即Single，表示单个字符
O，即Other，表示其他，用于标记无关字符

子集

IOB1: 标签I用于文本块中的字符，标签O用于文本块之外的字符，标签B用于在该文本块前面接续则一个同类型的文本块情况下的第一个字符。
IOB2: 每个文本块都以标签B开始，除此之外，跟IOB1一样。
IOE1: 标签I用于独立文本块中，标签E仅用于同类型文本块连续的情况，假如有两个同类型的文本块，那么标签E会被打在第一个文本块的最后一个字符。
IOE2: 每个文本块都以标签E结尾，无论该文本块有多少个字符，除此之外，跟IOE1一样。
START/END （也叫SBEIO、IOBES）: 包含了全部的5种标签，文本块由单个字符组成的时候，使用S标签来表示，由一个以上的字符组成时，首字符总是使用B标签，尾字符总是使用E标签，中间的字符使用I标签。
IO: 只使用I和O标签，显然，如果文本中有连续的同种类型实体的文本块，使用该标签方案不能够区分这种情况。
其中最常用的是IOB2、IOBS、IOBES。
"""
from itertools import chain
from typing import List

from ltp.core import Registrable


class Representation(metaclass=Registrable):
    def encode(self, text, tags=None):
        pass

    def decode(self, tags, text: List[str] = None):
        pass


# 这里暂时没有检查标签序列一致性
class IOBES(Representation):
    B: str = 'B'  # Begin        表示开始
    I: str = 'I'  # Intermediate 表示中间
    E: str = 'E'  # End          表示结尾
    S: str = 'S'  # Single       表示单个字符
    O: str = 'O'  # Other        表示其他，用于标记无关字符

    def encode(self, words: List[str], tags: List[str] = None):
        res = []
        # tags == None
        if tags is None:
            for word in words:
                word_len = len(word)
                if word_len == 1:
                    res.append(self.S)
                elif word_len == 2:
                    res.extend([self.B, self.E])
                else:
                    res.extend([self.B] + [self.I] * (word_len - 2) + [self.E])
        else:
            tags = ['-' + tag if tag != self.O else None for tag in tags]
            for word, tag in zip(words, tags):
                word_len = len(word)
                if tag is None:
                    res.extend([self.O] * word_len)
                elif word_len == 1:
                    res.append(self.S + tag)
                elif word_len == 2:
                    res.extend([self.B + tag, self.E + tag])
                else:
                    res.extend([self.B + tag] + [self.I + tag] * (word_len - 2) + [self.E + tag])

        return res

    def decode(self, tags: List[str], text: str = None):
        res_words = []
        res_tags = []

        if text == None:
            for tag in tags:
                iobes, i_tag = tag.split('-', 1)
                if iobes == self.I or iobes == self.E:
                    pass
                elif iobes == self.S or iobes == self.B:
                    res_tags.append(i_tag)
                else:  # Others
                    res_tags.append(iobes)
            return res_tags
        else:
            for tag, char in zip(tags, text):
                try:
                    iobes, i_tag = tag.split('-', 1)
                except Exception:
                    iobes = tag
                    i_tag = ''
                if iobes == self.I or iobes == self.E:
                    res_words[-1] += char
                elif iobes == self.S or iobes == self.B:
                    res_words.append(char)
                    res_tags.append(i_tag)
                else:  # Others
                    res_words.append(char)
                    res_tags.append(iobes)
            return res_words, res_tags


class IOB2(IOBES):
    def encode(self, words: List[str], tags: List[str] = None):
        res = []
        # tags == None
        if tags is None:
            return list(chain(*(["B"] + ["I"] * (len(word) - 1) for word in words)))
        else:
            tags = ['-' + tag if tag != self.O else None for tag in tags]
            for word, tag in zip(words, tags):
                word_len = len(word)
                if tag is None:
                    res.extend([self.O] * word_len)
                else:
                    res.extend([self.B + tag] + [self.I + tag] * (word_len - 1))
        return res



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>
import torch
from . import Predictor
from ltp.utils import length_to_mask


class Biaffine(Predictor, alias='biaffine'):
    def __init__(self, fields, text_field='text', tag_field='tag'):
        super().__init__(fields)
        self.text_field = text_field
        self.tag_vocab = self.fields[tag_field].vocab.itos

    def _convert_idx_to_name(self, y, array_len):
        return [[self.tag_vocab[idx] for idx in row[:row_len]] for row, row_len in zip(y, array_len)]

    def predict(self, inputs, pred):
        srl_output, srl_length = pred
        mask = length_to_mask(srl_length)

        mask = mask.unsqueeze_(-1).expand(-1, -1, mask.size(1))
        mask = (mask & mask.transpose(-1, -2)).flatten(end_dim=1)
        index = mask[:, 0]
        mask = mask[index]

        srl_output = srl_output.flatten(end_dim=1)[index]
        srl_labels = torch.argmax(srl_output, dim=-1).cpu().numpy()
        srl_labels = self._convert_idx_to_name(srl_labels, mask.sum(dim=1))
        # srl_labels_res = []
        # for length in srl_length:
        #     srl_labels_res.append([])
        #     curr_srl_labels, srl_labels = srl_labels[:length], srl_labels[length:]
        #     srl_labels_res[-1].extend([get_entities(labels) for labels in curr_srl_labels])

        return srl_labels



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>
from torch.nn.utils import clip_grad_norm_


def clip_grad_norm(parameters, max_norm, norm_type=None, pretrained_norm=None):
    if max_norm is None:
        return
    if norm_type is None:
        norm_type = 2
    if pretrained_norm is None:
        clip_grad_norm_(filter(lambda p: p.requires_grad, parameters), max_norm, norm_type)
    else:
        in_pretrained = []
        out_pretrained = []
        for n, p in parameters:
            if not p.requires_grad:
                continue
            if 'pretrained' in n:
                in_pretrained.append(p)
            else:
                out_pretrained.append(p)
        clip_grad_norm_(in_pretrained, max_norm, norm_type)
        clip_grad_norm_(out_pretrained, pretrained_norm, norm_type)



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>
from typing import Tuple, Union, Dict

import torch
from transformers import PretrainedConfig

from .model import Model


class MultiTaskModel(Model, alias='_multi'):
    """
    基本多任务序列标注模型
    """

    def create_decoder(self, input_size, label_num, dropout=0.1, **kwargs):
        """
        封装了各种解码器
        :param decoder: [None=Linear,lan=lan decoder,crf] 默认是简单线性层分类，目前支持 lan decoder

        :param hidden_size: decoder = lan 时必填, lan hidden size
        :param num_heads: decoder = lan 时使用, lan 多头注意力模型 heads，默认为 5
        :param num_layers: decoder = lan 时使用, lan decoder 层数，默认为 3
        :param lan: decoder = lan 时使用，lan decoder 其他参数

        :param arc_hidden_size: decoder = biaffine 时必填
        :param rel_hidden_size: decoder = biaffine 时必填
        :param rel_num: decoder = biaffine 时必填，rel 数目

        :param bias: decoder = Linear 时，传入Linear层
        """
        decoder_type = kwargs.pop('decoder', 'Linear')
        return super(MultiTaskModel, self).create_decoder(
            input_size, label_num, dropout, decoder=decoder_type, **kwargs
        )


class SimpleMultiTaskModel(MultiTaskModel):
    def __init__(self, pretrained: str = None, config: Union[str, PretrainedConfig] = None,
                 dropout=0.1, freeze=False, **kwargs):
        super().__init__()
        self.pretrained = self.create_pretrained(pretrained, config=config, freeze=freeze)
        config = self.pretrained.config
        self.emb_dropout = torch.nn.Dropout(p=dropout)

        self.task = None
        self.decoders_word_base = {}
        self.decoders_use_cls = {}
        for task, decoder_kwargs in kwargs.items():
            self.decoders_word_base[task] = decoder_kwargs.pop('word_base', False)
            self.decoders_use_cls[task] = decoder_kwargs.pop('use_cls', False)
            setattr(self, f"{task}_decoder", self.create_decoder(config.hidden_size, dropout=dropout, **decoder_kwargs))

    def __getitem__(self, item):
        self.task = item
        return self

    def decode(self, pretrained_output, rnn_steps, *args, **kwargs):
        decoder = getattr(self, f"{self.task}_decoder")
        if isinstance(decoder, torch.nn.Linear):
            return decoder(pretrained_output)
        else:
            return decoder(pretrained_output, rnn_steps, *args, **kwargs)

    def forward(self, text: Dict[str, torch.Tensor], *args, **kwargs):
        pretrained_output, *_ = self.pretrained(
            text['input_ids'],
            attention_mask=text['attention_mask'],
            token_type_ids=text['token_type_ids']
        )

        # remove [CLS] [SEP]
        use_cls = self.decoders_use_cls[self.task]
        pretrained_output = torch.narrow(
            pretrained_output, 1,
            1 - use_cls, pretrained_output.size(1) - 2 + use_cls
        )
        pretrained_output = self.emb_dropout(pretrained_output)

        if self.decoders_word_base[self.task]:
            word_idx, word_idx_len = text['word_index'], text['word_length']
            if use_cls:
                cls_tensor = torch.zeros((word_idx.shape[0], 1), dtype=word_idx.dtype, device=word_idx.device)
                word_idx = torch.cat([cls_tensor, word_idx + 1], dim=-1)
            word_idx = word_idx.unsqueeze(-1).expand(-1, -1, pretrained_output.shape[-1])
            pretrained_output = torch.gather(pretrained_output, dim=1, index=word_idx)
            seq_lens = word_idx_len
        else:
            seq_lens = text['text_length']

        return self.decode(pretrained_output, seq_lens, *args, **kwargs)



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>

from collections import Counter, OrderedDict
from itertools import chain
from typing import Union

import torch

from torchtext.vocab import Vocab
from ltp.const import PAD
from . import Field
from ltp.data.dataset import Dataset


def dtype_to_attr(dtype):
    # convert torch.dtype to dtype string id
    # e.g. torch.int32 -> "int32"
    # used for serialization
    _, dtype = str(dtype).split('.')
    return dtype


class SequenceField(Field, alias='sequence'):
    """
    序列 Field，通常为target


    :param name: Field name
    :param bos: Begin Of Sentence，默认为空
    :param eos: End Of Sentence，默认为空
    :param unk: Unknown Tag, 默认为空
    :param pad: 默认为 [PAD] 或 -1
    :param dtype: torch.dtype，可以使用字符串
    :param pad_bias: 做值域变换，将pad的值变到1，默认开启
    :param preprocessing: 预处理
    :param postprocessing: 后处理
    :param max_length: 是否padding到最大长度，None即为不做特殊处理，默认为None
    :param include_lengths: 是否返回 length，默认为False
    :param use_vocab: 是否使用词表，默认为True
    :param is_target: 是否为target，默认为True
    """

    vocab_cls = Vocab

    dtypes = {
        torch.float32: float,
        torch.float: float,
        torch.float64: float,
        torch.double: float,
        torch.float16: float,
        torch.half: float,

        torch.uint8: int,
        torch.int8: int,
        torch.int16: int,
        torch.short: int,
        torch.int32: int,
        torch.int: int,
        torch.int64: int,
        torch.long: int,
    }

    ignore = ['dtype']

    def __init__(self, name, bos: Union[str, int] = None, eos: Union[str, int] = None,
                 unk: Union[str, int] = None, pad: Union[str, int] = None,
                 dtype=torch.long, pad_bias=True, preprocessing=None, postprocessing=None,
                 max_length: int = None, include_lengths=False, labels=None,
                 use_vocab=True, is_target: bool = True, **kwargs):

        super(SequenceField, self).__init__(name, preprocessing, postprocessing, is_target)

        self.unk = unk if (isinstance(unk, str) and use_vocab) or (isinstance(unk, int) and not use_vocab) else None
        self.bos = bos if (isinstance(bos, str) and use_vocab) or (isinstance(bos, int) and not use_vocab) else None
        self.eos = eos if (isinstance(eos, str) and use_vocab) or (isinstance(eos, int) and not use_vocab) else None
        if use_vocab:
            self.pad = pad if isinstance(pad, str) else PAD
        elif not use_vocab:
            self.pad = pad if isinstance(pad, int) else -1

        if isinstance(dtype, str):
            self.dtype = getattr(torch, dtype)
        else:
            self.dtype = dtype
        self.use_vocab = use_vocab
        self.max_length = max_length
        self.include_lengths = include_lengths

        self.pad_bias = pad_bias

        if labels:
            counter = Counter()
            counter.update(labels)
            specials = list(
                OrderedDict.fromkeys(
                    tok for tok in [self.unk, self.pad, self.bos, self.eos] + kwargs.pop('specials', []) if
                    tok is not None)
            )
            self.vocab = self.vocab_cls(counter, specials=specials)

    def build_vocab(self, *args, **kwargs):
        if hasattr(self, 'vocab'):
            return
        counter = Counter()
        sources = []
        for arg in args:
            if isinstance(arg, Dataset):
                sources += [getattr(arg, name) for name, field in
                            arg.fields.items() if field is self]
            else:
                sources.append(arg)
        for data in sources:
            for x in data:
                try:
                    counter.update(x)
                except TypeError:
                    counter.update(chain.from_iterable(x))
        specials = list(
            OrderedDict.fromkeys(
                tok for tok in [self.unk, self.pad, self.bos, self.eos] + kwargs.pop('specials', []) if tok is not None)
        )
        self.vocab = self.vocab_cls(counter, specials=specials, **kwargs)

    def __setstate__(self, state):
        state['dtype'] = getattr(torch, state['dtype'])
        return super(SequenceField, self).__setstate__(state)

    def __getstate__(self):
        attrs = super(SequenceField, self).__getstate__()
        attrs['dtype'] = dtype_to_attr(self.dtype)
        return attrs

    def pad_batch(self, minibatch: list):
        minibatch = list(minibatch)
        max_len = max(len(x) for x in minibatch)

        if self.max_length is not None:
            max_len = min(max_len, self.max_length + (self.bos, self.eos).count(None) - 2)

        padded, lengths = [], []

        for x in minibatch:
            padded.append(
                ([] if self.bos is None else [self.bos])
                + list(x[:max_len])
                + ([] if self.eos is None else [self.eos])
                + [self.pad] * max(0, max_len - len(x[:max_len]))
            )
            lengths.append(len(padded[-1]) - max(0, max_len - len(x[:max_len])))

        if self.include_lengths:
            return (padded, lengths)
        return padded

    def numericalize(self, arr, device=None):
        if self.include_lengths and not isinstance(arr, tuple):
            raise ValueError("Field has include_lengths set to True, but "
                             "input data is not a tuple of "
                             "(data batch, batch lengths).")
        lengths = None
        if isinstance(arr, tuple):
            arr, lengths = arr

        if self.use_vocab:
            arr = [[self.vocab.stoi[x] for x in ex] for ex in arr]
            if self.postprocessing is not None:
                arr = self.postprocessing(arr, self.vocab)
        else:
            if self.dtype not in self.dtypes:
                raise ValueError(
                    f"Specified Field dtype {self.dtype} can not be used with "
                    "use_vocab=False because we do not know how to numericalize it. "
                )
            numericalization_func = self.dtypes[self.dtype]
            arr = [[numericalization_func(x) for x in ex] for ex in arr]
            if self.postprocessing is not None:
                arr = self.postprocessing(arr, None)

        var = torch.tensor(arr, dtype=self.dtype, device=device)
        var = var.contiguous()
        if self.include_lengths:
            lengths = torch.tensor(lengths, dtype=self.dtype, device=device)
            return var, lengths
        return var

    def preprocess(self, x):
        if self.preprocessing is not None:
            x = self.preprocessing(x)
        return x

    def process(self, batch, device=None):
        padded = self.pad_batch(batch)
        tensor = self.numericalize(padded, device=device)

        if not self.pad_bias:
            return tensor
        if isinstance(tensor, torch.Tensor):
            return tensor - 1
        else:
            tensor, length = tensor
            return tensor - 1, length



#!usr/bin/env python
# -*- coding:utf-8 -*-
# Author: jeffrey
# Changed:
#   sen -> sent / auto strip: ylfeng
import re
from typing import List


def split_sentence(document: str, flag: str = "all", limit: int = 510) -> List[str]:

    """

    Args:
        document:
        flag: Type:str, "all" 中英文标点分句，"zh" 中文标点分句，"en" 英文标点分句
        limit: 默认单句最大长度为510个字符

    Returns: Type:list

    """
    sent_list = []
    try:
        if flag == "zh":
            document = re.sub('(?P<quotation_mark>([。？！…](?![”’"\'])))', r'\g<quotation_mark>\n', document)  # 单字符断句符
            document = re.sub('(?P<quotation_mark>([。？！]|…{1,2})[”’"\'])', r'\g<quotation_mark>\n', document)  # 特殊引号
        elif flag == "en":
            document = re.sub('(?P<quotation_mark>([.?!](?![”’"\'])))', r'\g<quotation_mark>\n', document)  # 英文单字符断句符
            document = re.sub('(?P<quotation_mark>([?!.]["\']))', r'\g<quotation_mark>\n', document)  # 特殊引号
        else:
            document = re.sub('(?P<quotation_mark>([。？！….?!](?![”’"\'])))', r'\g<quotation_mark>\n', document)  # 单字符断句符
            document = re.sub('(?P<quotation_mark>(([。？！.!?]|…{1,2})[”’"\']))', r'\g<quotation_mark>\n',
                              document)  # 特殊引号

        sent_list_ori = document.splitlines()
        for sent in sent_list_ori:
            sent = sent.strip()
            if not sent:
                continue
            else:
                while len(sent) > limit:
                    temp = sent[0:limit]
                    sent_list.append(temp)
                    sent = sent[limit:]
                sent_list.append(sent)
    except:
        sent_list.clear()
        sent_list.append(document)
    return sent_list



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>
from typing import Optional
from torch import Tensor, nn

from ltp.modules import Module
from ltp import nn as ltp_nn


class RelativeTransformer(Module):
    def __init__(self, input_size: int, label_num: int, dropout: float, **kwargs):
        super().__init__(input_size, label_num, dropout)
        kwargs.setdefault('dropout', dropout)

        self.transformer = ltp_nn.RelativeTransformer(input_size, **kwargs)
        self.mlp = nn.Linear(input_size, label_num)

    def forward(self, inputs: Tensor, length: Tensor = None, gold: Optional = None):
        inputs, length, gold = self.transformer(inputs, length, gold)
        return self.mlp(inputs)



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>
from torch.optim.lr_scheduler import LambdaLR
from . import Scheduler


def StepExponentialLR(optimizer, decay_rate, decay_steps, staircase=True, last_epoch=-1):
    def lr_lambda(current_step):
        if staircase:
            return decay_rate ** (current_step // decay_steps)
        return decay_rate ** (current_step / decay_steps)

    return LambdaLR(optimizer, lr_lambda, last_epoch=last_epoch)


Scheduler.weak_register("StepExponentialLR", StepExponentialLR)



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>
from .task import TaskConfig, Task
from .config import Config
from .executor import Executor
from .commands import Command



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>
from typing import Optional

from torch import nn, Tensor
from . import Module


class Linear(nn.Linear):
    def __init__(self, input_size, label_num, dropout=None, bias=True):
        super().__init__(input_size, label_num, bias)

    def forward(self, embedding: Tensor, seq_lens: Tensor = None, gold: Optional = None):
        return super(Linear, self).forward(embedding)


Module.weak_register('Linear', Linear)



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>

from typing import Dict, Generic, TypeVar

import torch
from ltp.core import Registrable
from ltp.data import PreProcessing, PostProcessing


class Field(metaclass=Registrable):
    """ 通用基础 Field 类型

    每个数据集都会包含各种类型的数据，它们都可以被表示为Field，它描述了数据被处理的过程。

    在配置文件中配置项为::

        [[Fields]]
        class = "BertField"
        name = "text"
        [Fields.init]
        batch_first=true
        pretrained="data/albert_g/vocab.txt"
        config={"do_lower_case" = false}
        include_lengths=false

        [[Fields]]
        class = "TagField"
        name = "text_len"
        [Fields.init]
        is_target=false
        use_vocab=false

    属性:
        is_target: 是否是目标域。
            Default: False
    """

    ignore = []

    def __init__(self, name, preprocessing=None, postprocessing=None, is_target=False):
        self.name = name
        self.is_target = is_target
        self.preprocessing = preprocessing
        self.postprocessing = postprocessing

    def __eq__(self, other) -> bool:
        if isinstance(self, other.__class__):
            return self.__dict__ == other.__dict__
        return NotImplemented

    def preprocess(self, x):
        raise NotImplementedError()

    def process(self, batch, device=None, *args, **kwargs):
        raise NotImplementedError()

    def __getstate__(self):
        return {k: v for k, v in self.__dict__.items() if k not in self.ignore}

    def __setstate__(self, state):
        self.__dict__.update(state)

    def state_dict(self):
        return self.__getstate__()

    def load_state_dict(self, state):
        self.__setstate__(state)

    @classmethod
    def from_extra(cls, extra: dict, subcls=None):
        res = {}
        init = extra["config"]['init']
        if "preprocessing" in init:
            res["preprocessing"] = PreProcessing.from_params(init["preprocessing"])
        if "postprocessing" in init:
            res["postprocessing"] = PostProcessing.from_params(init["postprocessing"])
        return res


from .text import TextField, MixedTextField
from .label import LabelField
from .sequence import SequenceField
from .graph import GraphField
from .biaffine import BiaffineField

__all__ = ['Field', 'TextField', 'MixedTextField', 'LabelField', 'SequenceField', 'GraphField', 'BiaffineField']



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>
from typing import Union, List

default_float_rounding = 4


class EarlyStopping(object):
    """
    早停止
    """
    metric: Union[str, List[str]]
    patience: int
    float_rounding: int

    def __init__(self, patience: int = None, metric: Union[str, List[str]] = None, float_rounding: int = None):
        if patience is None:
            patience = 5
        if patience < 1:
            raise ValueError("Argument patience should be positive integer.")

        self.counter = 0
        self.best_score = None
        self.metric = 'loss' if metric is None else metric
        self.patience = patience
        self.float_rounding = default_float_rounding if float_rounding is None else float_rounding

    def better(self, metric):
        return round(metric, self.float_rounding) >= round(self.best_score, self.float_rounding)

    def __call__(self, state):
        if isinstance(self.metric, list):
            metric = sum([state.get(metric_name) for metric_name in self.metric])
        else:
            metric = state.get(self.metric)

        if self.best_score is None:
            self.best_score = metric
        elif self.better(metric):
            self.best_score = metric
            self.counter = 0
        else:
            self.counter += 1
            if self.counter >= self.patience:
                return True

        return False



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>
from . import Callback


class ValidationCallback(Callback, alias="_valid"):
    def __init__(self, data_loader, metrics, task, device=None, dtype=None, non_blocking=False):
        super(ValidationCallback, self).__init__(None, 1)
        self.task = task
        self.preffix = '' if task == 'default' else (task + '_')
        self.data_loader = data_loader
        self.metrics = metrics
        self.device = device
        self.dtype = dtype
        self.non_blocking = non_blocking

    def init(self, executor):
        for metric in self.metrics:
            for name, value in metric.defaults.items():
                executor.trainer.state.add_attribute(self.preffix + name, value)
                # executor.add_progressbar_metric('dev_' + name)

    def call(self, executor):
        metrics = executor.evaluate_(self.data_loader, self.metrics, self.task)
        for metric in metrics:
            for attr, value in metric.compute().items():
                executor.trainer.state.update_attribute(self.preffix + attr, value)
            metric.clear()



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>
from typing import Optional
from torch import Tensor
from ltp.nn import MLP, Bilinear

from . import Module


class Graph(Module):
    def __init__(self, input_size, label_num, dropout, arc_hidden_size, rel_hidden_size, **kwargs):
        super(Graph, self).__init__(input_size, label_num, dropout)
        arc_dropout = kwargs.pop('arc_dropout', dropout)
        rel_dropout = kwargs.pop('rel_dropout', dropout)
        activation = kwargs.pop('activation', {})
        self.mlp_arc_h = MLP(input_size, arc_hidden_size, arc_dropout, **activation)
        self.mlp_arc_d = MLP(input_size, arc_hidden_size, arc_dropout, **activation)
        self.mlp_rel_h = MLP(input_size, rel_hidden_size, rel_dropout, **activation)
        self.mlp_rel_d = MLP(input_size, rel_hidden_size, rel_dropout, **activation)

        self.arc_atten = Bilinear(arc_hidden_size, arc_hidden_size, 1, bias_x=True, bias_y=False, expand=True)
        self.rel_atten = Bilinear(rel_hidden_size, rel_hidden_size, label_num, bias_x=True, bias_y=True, expand=True)

    def forward(self, embedding: Tensor, seq_lens: Tensor = None, gold: Optional = None):
        arc_h = self.mlp_arc_h(embedding)
        arc_d = self.mlp_arc_d(embedding)

        rel_h = self.mlp_rel_h(embedding)
        rel_d = self.mlp_rel_d(embedding)

        s_arc = self.arc_atten(arc_d, arc_h).squeeze_(1)
        s_rel = self.rel_atten(rel_d, rel_h).permute(0, 2, 3, 1)

        return s_arc, s_rel, seq_lens



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>
import os
import torch
import numpy as np
from ltp import LTP
from ltp.ltp import no_gard, TensorType


def convert(item: list):
    array = np.asarray(item, dtype=np.int64)

    if len(array.shape) == 1:
        array = np.expand_dims(array, 0)
    return array


class FastLTP(LTP):
    tensor = TensorType.NUMPY

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        import onnxruntime as rt
        so = rt.SessionOptions()
        so.graph_optimization_level = rt.GraphOptimizationLevel.ORT_ENABLE_EXTENDED

        # fixme should auto detect
        providers = ['CPUExecutionProvider'] if self.device.type == 'cpu' else ['GPUExecutionProvider']

        onnx_path = os.path.join(self.cache_dir, "ltp.onnx")
        if not os.path.isfile(onnx_path):
            ltp_onnx_export(self, onnx_path)

        self.onnx = rt.InferenceSession(onnx_path, so, providers=providers)

    def __str__(self):
        return f"FastLTP {self.version} on {self.device}"

    def __repr__(self):
        return f"FastLTP {self.version} on {self.device}"

    @no_gard
    def _seg(self, tokenizerd, *args, **kwargs):
        pretrained_inputs = {key: convert(value) for key, value in tokenizerd.items()}
        length = np.sum(pretrained_inputs['attention_mask'], axis=-1) - 2

        # todo: io binding
        cls, hidden, seg = self.onnx.run(None, pretrained_inputs)

        word_cls = torch.as_tensor(cls, device=self.device)
        char_input = torch.as_tensor(hidden, device=self.device)
        return word_cls, char_input, seg, length


def ltp_onnx_export(ltp: LTP, path: str):
    from torch.onnx import export

    dummy_input = {
        'input_ids': torch.as_tensor([
            [101, 800, 1373, 3739, 1990, 1343, 2897, 1912, 6132, 511, 102, 0, 0],
            [101, 2571, 3647, 4638, 1383, 2094, 1355, 7942, 5445, 1318, 3289, 511, 102]
        ], device=ltp.device),
        'token_type_ids': torch.as_tensor([
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        ], device=ltp.device),
        'attention_mask': torch.as_tensor([
            [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0],  # 11
            [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1]  # 13
        ], device=ltp.device),
    }

    input_names = ['input_ids', 'attention_mask', 'token_type_ids']
    output_names = ['hidden', 'seg']
    dynamic_axes = {
        'input_ids': {0: 'batch', 1: 'length'},
        'attention_mask': {0: 'batch', 1: 'length'},
        'token_type_ids': {0: 'batch', 1: 'length'},
        'hidden': {0: 'batch', 1: 'length'},
        'seg': {0: 'batch', 1: 'length'},
    }
    model_args = tuple(dummy_input[arg] for arg in input_names)

    class Model(torch.nn.Module):
        def __init__(self, pretrained, seg):
            super().__init__()
            self.pretrained = pretrained
            self.seg = seg

        def forward(self, *args, **kwargs):
            hidden = self.pretrained(*args, **kwargs)[0]

            cls = hidden[:, :1]
            hidden_cut = hidden[:, 1:-1]
            seg = self.seg(hidden_cut)
            seg = torch.argmax(seg, dim=-1)

            return cls, hidden_cut, seg

    model = Model(
        ltp.model.pretrained,
        ltp.model.seg_decoder
    )

    with torch.no_grad():
        export(
            model,
            model_args,
            f=path,
            input_names=input_names,
            output_names=output_names,
            dynamic_axes=dynamic_axes,
            do_constant_folding=True,
            use_external_data_format=False,
            enable_onnx_checker=True,
            opset_version=12,
        )



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>
"""
模型在配置文件中的配置项为::
    [Model]
    class = "SrlTagging"

        [Model.init]
        label_num=77
        hidden_size=400
        pretrained="data/bert_wwm"
"""

from .model import Model
from .seq_tag_model import SequenceTaggingModel
from .multi_task_model import MultiTaskModel, SimpleMultiTaskModel



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>
__version__ = '4.0.9'

from .core import Registrable
from .data import Dataset
from .eval import Metric
from .models import Model
from .predict import Predictor
from .train import Callback, Loss, Optimizer, Scheduler, Trainer
from .exe import Executor, Command
from .ltp import LTP
from .fastltp import FastLTP



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>
from typing import Callable, Union
from . import Optimizer


def PretrainedOptim(named_params, optim: Union[str, Callable], lr=1e-5, weight_decay=0.01,
                    bert_lr=None, bert_weight_decay=None, **kwargs):
    if bert_lr is None:
        bert_lr = lr
    if bert_weight_decay is None:
        bert_weight_decay = weight_decay
    param_optimizer = list(named_params)
    no_decay = ['bias', 'LayerNorm.bias', 'LayerNorm.weight']
    params = {'in_pretrained': {'decay': [], 'no_decay': []}, 'out_pretrained': {'decay': [], 'no_decay': []}}
    for n, p in param_optimizer:
        is_in_pretrained = 'in_pretrained' if 'pretrained' in n else 'out_pretrained'
        is_no_decay = 'no_decay' if any(nd in n for nd in no_decay) else 'decay'
        params[is_in_pretrained][is_no_decay].append(p)

    optimizer_grouped_parameters = [
        {'params': params['in_pretrained']['decay'], 'weight_decay': bert_weight_decay, 'lr': bert_lr},
        {'params': params['in_pretrained']['no_decay'], 'weight_decay': 0.0, 'lr': bert_lr},
        {'params': params['out_pretrained']['decay'], 'weight_decay': weight_decay, 'lr': lr},
        {'params': params['out_pretrained']['no_decay'], 'weight_decay': weight_decay, 'lr': lr},
    ]

    assert optim != "PretrainedOptim" and optim != "BertAdamW"

    if isinstance(optim, str):
        return Optimizer.by_name(optim)(
            optimizer_grouped_parameters, lr=lr,
            weight_decay=weight_decay, **kwargs)
    elif isinstance(optim, Callable):
        return optim(
            optimizer_grouped_parameters, lr=lr,
            weight_decay=weight_decay, **kwargs)
    else:
        raise NotImplementedError(f"Optim not support {type(optim)}")


def BertAdamW(named_params, lr=1e-5, weight_decay=0.01, eps=1e-8, **kwargs):
    from transformers import optimization
    return PretrainedOptim(named_params, lr=lr, weight_decay=weight_decay, optim=optimization.AdamW, eps=eps, **kwargs)


Optimizer.weak_register("PretrainedOptim", PretrainedOptim)
Optimizer.weak_register("BertAdamW", BertAdamW)



from typing import List, Dict
from . import Dataset


class MixedDataset(Dataset, alias='mixed'):
    def __init__(self, path: List[Dict], file, fields, text='text'):
        datasets = {}
        if isinstance(file, str):
            file = [file] * len(path)
        for dataset, file in zip(path, file):
            init = {}
            name = dataset['name']
            for key, value in dataset.items():
                if key != 'name' and key != 'path':
                    init[key] = value
            datasets[name] = Dataset.from_params(init, path=dataset['path'], file=file, fields=fields)

        examples = []
        for name, dataset in datasets.items():
            for example in dataset.examples:
                setattr(example, text, (name, *getattr(example, text)))
                examples.append(example)

        super().__init__(examples, fields)



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>
from functools import partial
import torch
from torch import Tensor
from . import Trainer
from ltp.utils import clip_grad_norm, select_logits_with_mask
from ltp.predict import Predictor
from ltp.train.scheduler import TemperatureScheduler
from ltp.train.scheduler.temperature_scheduler import ConstantScheduler
from ltp.train.loss import Loss


class MultiTaskDistiller(Trainer, alias='multi_dist'):
    def __init__(self, config, temperature=8, kd_loss_weight=1, hard_label_weight=1):
        super().__init__(config)
        self.verbose = self.config.verbose
        self.tasks = self.config.tasks
        self.task = self.tasks['default']
        self.model = self.task.model

        self.temperature = temperature
        self._kd_loss_weight = kd_loss_weight
        self._hard_label_weight = hard_label_weight

        if hasattr(self.config, 'temperature_scheduler'):
            self.temperature_scheduler = TemperatureScheduler.from_params(self.config.temperature_scheduler)
        else:
            self.temperature_scheduler = ConstantScheduler()

        if hasattr(self.model, 'extractor'):
            self.data_extractor = lambda batch: self.model.extractor(batch)
        else:
            self.data_extractor = self.__data_extractor
        if self.verbose:
            print(self.model)

    @staticmethod
    def __data_extractor(batch, task):
        x = tuple(getattr(batch, name) for name, _ in task.fields if name in batch.input_fields)
        y = tuple(getattr(batch, name) for name, _ in task.fields if name in batch.target_fields)
        return x, y

    @property
    def kd_loss_weight(self) -> float:
        return self._kd_loss_weight * (1 - self.state.global_step / self.total_steps)

    @property
    def hard_label_weight(self) -> float:
        return self._hard_label_weight * (self.state.global_step / self.total_steps)

    def train(self, batch, task):
        self.task.optimizer.zero_grad()
        x, y = self.data_extractor(batch, self.tasks[task])

        if len(y) == 0:
            y = x
        if len(y) == 1:
            y = y[0]

        if isinstance(x, tuple):
            with torch.no_grad():
                logits_T = self.tasks[task].model(*x)
            logits = self.model[task](*x)
        elif isinstance(x, dict):
            with torch.no_grad():
                logits_T = self.tasks[task].model(**x)
            logits = self.model[task](**x)
        else:
            raise NotImplementedError("Model.extrator 的返回值 x 必须是一个tuple或者一个 dict")

        # 使用每个任务的 loss 计算 hard loss
        if isinstance(self.tasks[task].loss, Loss):
            # Hard Label Loss
            hard_label_loss = self.tasks[task].loss(logits, y)

            # Distill Loss
            temperature_calc = partial(self.temperature_scheduler, base_temperature=self.temperature)
            kd_loss = self.tasks[task].loss.distill(logits, logits_T, temperature_calc, self.task.loss, gold=y)
        elif isinstance(y, Tensor):
            # Hard Label Loss
            shape = logits.shape[-1]
            hard_label_loss = self.tasks[task].loss(logits.contiguous().view((-1, shape)), y.contiguous().view(-1))

            # Distill Loss
            mask = torch.ne(y, -1)
            logits = select_logits_with_mask(logits, mask)
            logits_T = select_logits_with_mask(logits_T, mask)

            temperature = self.temperature_scheduler(logits, logits_T, self.temperature)
            kd_loss = self.task.loss(logits, logits_T, temperature)
        else:
            raise NotImplementedError("多输出结果需要自行实现Loss")

        loss = kd_loss * self.kd_loss_weight + hard_label_loss * self.hard_label_weight
        loss.backward()

        clip_grad_norm(
            self.model.named_parameters() if self.config.pretrained_grad_norm else self.model.parameters(),
            self.config.grad_norm, norm_type=self.config.norm_type, pretrained_norm=self.config.pretrained_grad_norm
        )

        self.task.optimizer.step()
        if self.task.scheduler_type == 'step' and self.task.scheduler:
            self.task.scheduler.step()
        return loss.item()

    def after_train(self):
        if self.task.scheduler_type == 'epoch' and self.task.scheduler:
            self.task.scheduler.step()
        self.task.save(self.state)

    def before_eval(self, task):
        self.previous_training_flag = self.model.training
        self.model[task].eval()

    def eval(self, batch, task):
        x, y = self.data_extractor(batch, self.tasks[task])

        if len(y) == 0:
            y = x
        if len(y) == 1:
            y = y[0]

        if isinstance(x, tuple):
            y_pred = self.model(*x)
        elif isinstance(x, dict):
            y_pred = self.model(**x)
        else:
            raise NotImplementedError("Model.extrator 的返回值 x 必须是一个tuple或者一个 dict")

        return x, y_pred, y

    def after_eval(self, task):
        self.model.train(self.previous_training_flag)

    def before_predict(self, task):
        self.previous_training_flag = self.model.training
        self.model[task].eval()
        if self.tasks[task].config.predictor:
            self.predictor = Predictor.from_params(
                self.tasks[task].config.predictor,
                fields=self.tasks[task].fields
            )
        else:
            print("No Predictor")

    def predict(self, batch, task):
        x, y = self.data_extractor(batch, self.tasks[task])

        if isinstance(x, tuple):
            y_pred = self.model(*x)
        elif isinstance(x, dict):
            y_pred = self.model(**x)
        else:
            raise NotImplementedError("Model.extrator 的返回值 x 必须是一个tuple或者一个 dict")

        return self.predictor(batch, y_pred)

    def after_predict(self, task):
        self.model.train(self.previous_training_flag)



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>

import codecs

def iter_raw_lines(filename: str, strip=None, skip: str = '#'):
    line_num = 0
    with codecs.open(filename, encoding='utf-8') as file:
        while True:
            line = file.readline()
            line_num += 1
            if line.startswith(skip):
                continue
            if not line:  # EOF
                yield line_num, ''  # 输出空行，简化上层逻辑
                break
            line = line.strip(strip)
            yield line_num, line


def iter_lines(filename: str, split=None, strip=None):
    for line_num, raw_line in iter_raw_lines(filename=filename, strip=strip):
        if not raw_line:  # end of a sentence
            yield line_num, []  # 输出空行
        else:
            yield line_num, raw_line.split(split)


def iter_blocks(filename: str, split=None, strip=None):
    rows = []
    for line_num, line_features in iter_lines(filename, split=split, strip=strip):
        if len(line_features):
            rows.append(line_features)
        else:
            if len(rows):
                yield line_num, rows
                rows = []



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>
from typing import Tuple, Union, Dict, Optional

import torch
from transformers import AutoConfig, AutoModel, PretrainedConfig

from .model import Model
from ltp.modules import Module


class SequenceTaggingModel(Model, alias="_sequence_tagging"):
    """
    基本序列标注模型，封装了各种解码器，将来可能要进行解耦，如果使用crf解码器，需要换用支持 CRF 的 Trainer

    使用 init_decoder 函数初始化解码器

    :param decoder: [None=Linear,lan=lan decoder] 默认是简单线性层分类，目前支持 lan decoder

    :param hidden_size: decoder = lan 时必填, lan hidden size
    :param num_heads: decoder = lan 时使用, lan 多头注意力模型 heads，默认为 5
    :param num_layers: decoder = lan 时使用, lan decoder 层数，默认为 3
    :param lan: decoder = lan 时使用，lan decoder 其他参数

    :param arc_hidden_size: decoder = biaffine 时必填
    :param rel_hidden_size: decoder = biaffine 时必填
    :param rel_num: decoder = biaffine 时默认为label_num，rel 数目

    :param bias: decoder = Linear 时，传入Linear层
    """

    decoder: Module

    def init_decoder(self, input_size, label_num, dropout=0.1, **kwargs):
        """
        基本序列标注模型，封装了各种解码器
        """
        self.decoder = self.create_decoder(input_size, label_num, dropout, **kwargs)

    def hidden2logits(self, logits, seq_len, gold: Optional = None):
        return self.decoder.forward(logits, seq_len, gold)

    def forward_(self, *inputs, **kwargs) -> Tuple[torch.Tensor, torch.Tensor]:
        raise NotImplementedError("未实现")

    def forward(self, text: Dict[str, torch.Tensor], gold: Optional = None):
        features, seq_lens = self.forward_(text)
        return self.hidden2logits(features, seq_lens, gold)


class SimpleTaggingModel(SequenceTaggingModel, alias="SimpleTagging"):
    """
    基本序列标注模型

    :param pretrained: 预训练模型路径或名称，参照 huggingface/transformers
    :param config: 预训练模型路径或名称，参照 huggingface/transformers
    :param label_num: 分类标签数目
    :param dropout: pretrained Embedding dropout，默认0.1
    :param word_base: 是否是以词为基础的模型，如果是，输入时需要传入 word index
    :param decoder: [None=Linear,lan=lan decoder] 默认是简单线性层分类，目前支持 lan decoder

    :param hidden_size: decoder = lan 时必填, lan hidden size
    :param num_heads: decoder = lan 时使用, lan 多头注意力模型 heads，默认为 5
    :param num_layers: decoder = lan 时使用, lan decoder 层数,，默认为 3
    :param lan: decoder = lan 时使用，lan decoder 其他参数

    :param arc_hidden_size: decoder = graph 时必填
    :param rel_hidden_size: decoder = graph 时必填
    :param rel_num: decoder = graph 时默认为label_num，rel 数目

    :param bias: decoder = Linear 时，传入Linear层
    """

    def __init__(self, label_num, pretrained: str = None, config: Union[str, PretrainedConfig] = None,
                 dropout=0.1, word_base=False, use_cls=False, freeze=False, **kwargs):
        """
        基本序列标注模型
        """
        super(SimpleTaggingModel, self).__init__()
        self.word_base = word_base
        self.use_cls = use_cls
        self.gard_enable = not freeze
        self.pretrained = self.create_pretrained(pretrained, config=config, freeze=freeze)
        config = self.pretrained.config
        self.emb_dropout = torch.nn.Dropout(p=dropout)

        self.init_decoder(
            input_size=config.hidden_size,
            label_num=label_num,
            dropout=dropout,
            **kwargs
        )

    def forward_(self, text: Dict[str, torch.Tensor]):
        with torch.set_grad_enabled(self.gard_enable):
            pretrained_output, *_ = self.pretrained(
                text['input_ids'],
                attention_mask=text['attention_mask'],
                token_type_ids=text['token_type_ids']
            )
            # remove [CLS] [SEP]
            pretrained_output = torch.narrow(
                pretrained_output, 1,
                1 - self.use_cls, pretrained_output.size(1) - 2 + self.use_cls
            )
            pretrained_output = self.emb_dropout(pretrained_output)
            if self.word_base:
                word_idx = text['word_index']
                if self.use_cls:
                    cls_tensor = torch.zeros((word_idx.shape[0], 1), dtype=word_idx.dtype, device=word_idx.device)
                    word_idx = torch.cat([cls_tensor, word_idx + 1], dim=-1)
                word_idx = word_idx.unsqueeze(-1).expand(-1, -1, pretrained_output.shape[-1])
                pretrained_output = torch.gather(pretrained_output, dim=1, index=word_idx)
                seq_lens = text['word_length']
            else:
                seq_lens = text['text_length']
            return pretrained_output, seq_lens



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>

import six
import json
from functools import reduce


class Example(object):
    @classmethod
    def fromJSON(cls, data, fields):
        ex = cls()
        obj = json.loads(data)

        for key, vals in fields.items():
            if vals is not None:
                if not isinstance(vals, list):
                    vals = [vals]

                for val in vals:
                    # for processing the key likes 'foo.bar'
                    name, field = val
                    ks = key.split('.')

                    def reducer(obj, key):
                        if isinstance(obj, list):
                            results = []
                            for data in obj:
                                if key not in data:
                                    # key error
                                    raise ValueError("Specified key {} was not found in "
                                                     "the input data".format(key))
                                else:
                                    results.append(data[key])
                            return results
                        else:
                            # key error
                            if key not in obj:
                                raise ValueError("Specified key {} was not found in "
                                                 "the input data".format(key))
                            else:
                                return obj[key]

                    v = reduce(reducer, ks, obj)
                    setattr(ex, name, field.preprocess(v))
        return ex

    @classmethod
    def fromdict(cls, data, fields):
        ex = cls()
        for key, vals in fields.items():
            if key not in data:
                raise ValueError("Specified key {} was not found in "
                                 "the input data".format(key))
            if vals is not None:
                if not isinstance(vals, list):
                    vals = [vals]
                for val in vals:
                    name, field = val
                    setattr(ex, name, field.preprocess(data[key]))
        return ex

    @classmethod
    def fromCSV(cls, data, fields, field_to_index=None):
        if field_to_index is None:
            return cls.fromlist(data, fields)
        else:
            assert (isinstance(fields, dict))
            data_dict = {f: data[idx] for f, idx in field_to_index.items()}
            return cls.fromdict(data_dict, fields)

    @classmethod
    def fromlist(cls, data, fields):
        ex = cls()
        for (name, field), val in zip(fields, data):
            if field is not None:
                if isinstance(val, six.string_types):
                    val = val.rstrip('\n')
                # Handle field tuples
                if isinstance(name, tuple):
                    for n, f in zip(name, field):
                        setattr(ex, n, f.preprocess(val))
                else:
                    setattr(ex, name, field.preprocess(val))
        return ex

    @classmethod
    def fromtree(cls, data, fields, subtrees=False):
        try:
            from nltk.tree import Tree
        except ImportError:
            print("Please install NLTK. "
                  "See the docs at http://nltk.org for more information.")
            raise
        tree = Tree.fromstring(data)
        if subtrees:
            return [cls.fromlist(
                [' '.join(t.leaves()), t.label()], fields) for t in tree.subtrees()]
        return cls.fromlist([' '.join(tree.leaves()), tree.label()], fields)



from . import Sequence
from ltp.utils import length_to_mask


class Segment(Sequence):
    def __init__(self, *args):
        super(Segment, self).__init__(['B-W', 'I-W'])

    def step(self, y_pred, y: dict):
        mask = ~ length_to_mask(y['text_length'])
        target = y['word_idn']
        target[mask] = -1
        super(Segment, self).step(y_pred, target)

    @classmethod
    def from_extra(cls, extra: dict):
        return {}



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>
from torch import Tensor
from torch.nn.modules.loss import _Loss as Loss
from . import Metric


class CommonLoss(Metric, alias="Loss"):
    """
    Loss Metric: 用于计算LOSS

    :param loss_function: 损失函数
    :param item: 返回标量，而不是Tensor
    :param flat: 是否对预测结果进行平铺
    """

    def __init__(self, loss_function: Loss, item: bool = False, flat: bool = False):
        """
        Loss Metric: 用于计算LOSS

        :param loss_function: 损失函数
        :param item: 返回标量，而不是Tensor
        :param flat: 是否对预测结果进行平铺
        """

        super(CommonLoss, self).__init__(loss=float('inf'))
        self._total = 0
        self._flat = flat
        self._loss_sum = 0.
        self._return_item = item
        self.loss_function = loss_function

    def step(self, y_pred: Tensor, y: Tensor):
        if self._flat:
            shape = y_pred.shape[-1]
            y_pred, y = y_pred.contiguous().view((-1, shape)), y.contiguous().view(-1)
        loss = self.loss_function(y_pred, y)
        self._loss_sum += loss.item()
        self._total += 1

    def compute(self):
        if self._total == 0:
            raise ZeroDivisionError("Loss average is not computable.")
        return {'loss': (self._loss_sum / self._total) if self._total != 0 else 0}

    def clear(self):
        self._loss_sum = 0.
        self._total = 0



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>
from torch import nn
from ltp.core import Registrable


class Module(nn.Module, metaclass=Registrable):
    def __init__(self, input_size, label_num, dropout):
        super(Module, self).__init__()
        self.input_size = input_size
        self.label_num = label_num
        self.dropout = dropout

    def forward(self, embedding, seq_lens, gold=None):
        raise NotImplementedError()


from .linear import Linear
from .graph import Graph
from .biaffine_crf import BiaffineCRF
from .bilinear_crf import BiLinearCRF
from .relative_transformer import RelativeTransformer



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>
import torch
from . import Predictor


class Simple(Predictor, alias='simple'):
    """
    简单预测器

    :param fields: Tuple(str, Field) 数据的 Field 对象，配置文件不需要填写
    :param text_field: str 提供文本以及长度的 Field，默认是第一个 Filed
    :param target_filed: str 提供解码词典的 Field, 默认是第一个 Target Field (提供他的词表)
    """

    def __init__(self, fields, text_field: str = 'text', target_filed: str = None):
        """
        简单预测器
        """
        super(Simple, self).__init__(fields)
        self.text_field = self.fields[text_field]
        if target_filed is None:
            label_field = self.target_fields[0]
        else:
            label_field = self.fields.get(target_filed, default=self.target_fields[0])

        if label_field.use_vocab:
            self.itos = label_field.vocab.itos[1:]

    def convert_idx_to_name(self, y, array_len, id2label):
        if id2label:
            return [[id2label[idx] for idx in row[:row_len]] for row, row_len in zip(y, array_len)]
        else:
            return [[idx for idx in row[:row_len]] for row, row_len in zip(y, array_len)]

    def predict(self, inputs, pred):
        text_field = getattr(inputs, self.text_field)
        target_len = text_field[1] if isinstance(text_field, tuple) else text_field
        target_len = target_len.cpu().detach().numpy().tolist()

        pred = torch.argmax(pred, dim=-1).cpu().detach().numpy()
        pred = self.convert_idx_to_name(pred, target_len, self.itos)
        return pred



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>

import torch
from . import Metric


class Accuracy(Metric):
    def __init__(self, pad_value: int = None):
        """
        用于词性标注
        :param pad_value: 被忽略的目标值
        """
        super(Accuracy, self).__init__(acc=0.)
        self._total = 0
        self._pad_value = pad_value
        self._total_correct = 0

    def step(self, y_pred: torch.Tensor, y: torch.Tensor):
        y_pred = torch.argmax(y_pred, dim=-1)

        if y.size() != y_pred.size():
            raise TypeError("y and y_pred should have the same shape")
        correct = torch.eq(y_pred, y)

        mask = torch.ones_like(y, dtype=torch.bool) if self._pad_value is None else y.ne(self._pad_value)
        self._total_correct += torch.sum(correct[mask]).item()
        self._total += mask.sum().item()

    def compute(self):
        return {'acc': (self._total_correct / self._total) if self._total != 0 else 0}

    def clear(self):
        self._total_correct = 0
        self._total = 0



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>

from functools import partial
import os
import logging

import torch
from tqdm import tqdm

from torchtext.vocab import Vectors
from torchtext.vocab import Vocab
from torchtext.vocab import pretrained_aliases

logger = logging.getLogger(__name__)

backup_load_vectors = Vocab.load_vectors


def load_vectors(self, vectors, **kwargs):
    if not isinstance(vectors, list):
        vectors = [vectors]
    for vector in vectors:
        if vector not in pretrained_aliases:
            pretrained_aliases[vector] = partial(TextVectors, name=vector)

    backup_load_vectors(self, vectors, **kwargs)


Vocab.load_vectors = load_vectors


class TextVectors(Vectors):

    def __init__(self, name, cache=None, unk_init=None, max_vectors=None):
        super(TextVectors, self).__init__(name=name, cache=cache, unk_init=unk_init, max_vectors=max_vectors)

    def cache(self, name, cache, url=None, max_vectors=None):
        if os.path.isfile(name):
            path = name
            if max_vectors:
                file_suffix = '_{}.pt'.format(max_vectors)
            else:
                file_suffix = '.pt'
            path_pt = os.path.join(cache, os.path.basename(name)) + file_suffix
        else:
            raise FileNotFoundError(name)

        if not os.path.isfile(path_pt):
            vectors_loaded = 0
            with open(path, 'r') as f:
                itos, vectors, dim = [], [], None

                for line in tqdm(f, total=max_vectors):
                    # Explicitly splitting on " " is important, so we don't
                    # get rid of Unicode non-breaking spaces in the vectors.
                    entries = line.rstrip().split()

                    word, entries = entries[0], entries[1:]
                    if dim is None and len(entries) > 1:
                        dim = len(entries)
                    elif len(entries) == 1:
                        logger.warning("Skipping token {} with 1-dimensional "
                                       "vector {}; likely a header".format(word, entries))
                        continue
                    elif dim != len(entries):
                        raise RuntimeError(
                            "Vector for token {} has {} dimensions, but previously "
                            "read vectors have {} dimensions. All vectors must have "
                            "the same number of dimensions.".format(word, len(entries), dim))

                    vectors.append([float(x) for x in entries])
                    vectors_loaded += 1
                    itos.append(word)

                    if vectors_loaded == max_vectors:
                        break

            self.itos = itos
            self.stoi = {word: i for i, word in enumerate(itos)}
            self.vectors = torch.Tensor(vectors).contiguous().view(-1, dim)
            self.dim = dim
            logger.info('Saving vectors to {}'.format(path_pt))
            if not os.path.exists(cache):
                os.makedirs(cache)
            torch.save((self.itos, self.stoi, self.vectors, self.dim), path_pt)
        else:
            logger.info('Loading vectors from {}'.format(path_pt))
            self.itos, self.stoi, self.vectors, self.dim = torch.load(path_pt)



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>

import torch.utils.data
import random
from contextlib import contextmanager
from copy import deepcopy
import os

from ltp import Registrable


class Dataset(torch.utils.data.dataset.Dataset, metaclass=Registrable):
    """
    数据集抽象类，在配置文件中的配置项通常为::

        [Dataset]
        class = "CTB"

        [Dataset.init]
        path="data/postag"
        train="train.txt"
        validation="dev.txt"
        test="test.txt"

    定义一个由 Examples 和 Fields 组合成的 Dataset

    Attributes:
        sort_key (callable): 用于在生成Batch的是时候进行排序的函数
        examples (List[Example]): 数据集的 Examples
        fields (dict[str, Field]): Fields，同一个 Field Object 会共享它们的 Vocab
    """
    sort_key = None

    def __init__(self, examples, fields, filter_pred=None):
        """使用examples和fields创建一个Dataset

        Args:
            examples: Examples的列表
            fields: List(tuple(str, Field))
            filter_pred: 只有当该函数返回值为True时，该Example才会被使用
        """

        if filter_pred is not None:
            make_list = isinstance(examples, list)
            examples = filter(filter_pred, examples)
            if make_list:
                examples = list(examples)
        self.examples = examples
        self.fields = dict(fields)
        # Unpack field tuples
        for n, f in list(self.fields.items()):
            if isinstance(n, tuple):
                self.fields.update(zip(n, f))
                del self.fields[n]

    @classmethod
    def splits(cls, path=None, root='.data', train=None, validation=None,
               test=None, **kwargs):
        """ 一次创建多个数据集

        Args:
            path: 数据集路径前缀
            root: 根目录
            train: 训练集文件名
            validation: 验证集文件名
            test: 测试集文件名
            **kwargs: 其他传给Dataset的参数

        Returns:
            Tuple[Dataset] train, validation, test
        """
        if path is None:
            path = cls.download(root)
        train_data = None if train is None else cls(
            os.path.join(path, train), **kwargs)
        val_data = None if validation is None else cls(
            os.path.join(path, validation), **kwargs)
        test_data = None if test is None else cls(
            os.path.join(path, test), **kwargs)
        return tuple(d for d in (train_data, val_data, test_data)
                     if d is not None)

    def split(self, split_ratio=0.7, stratified=False, strata_field='label',
              random_state=None):
        """通过切分当前数据集的Examples以获得多个数据集

        Args:
            split_ratio (float/list(float)): 切分比例(默认为0.7)
            stratified: 是否进行分层采样
            strata_field: 分层采样的field
            random_state: 用于Shuffle的随机种子`random.getstate()`的返回值

        Returns:
            Tuple[Dataset] train, validation, test Dataset
        """
        train_ratio, test_ratio, val_ratio = check_split_ratio(split_ratio)

        # For the permutations
        rnd = RandomShuffler(random_state)
        if not stratified:
            train_data, test_data, val_data = rationed_split(self.examples, train_ratio,
                                                             test_ratio, val_ratio, rnd)
        else:
            if strata_field not in self.fields:
                raise ValueError("Invalid field name for strata_field {}"
                                 .format(strata_field))
            strata = stratify(self.examples, strata_field)
            train_data, test_data, val_data = [], [], []
            for group in strata:
                # Stratify each group and add together the indices.
                group_train, group_test, group_val = rationed_split(group, train_ratio,
                                                                    test_ratio, val_ratio,
                                                                    rnd)
                train_data += group_train
                test_data += group_test
                val_data += group_val

        splits = tuple(Dataset(d, self.fields)
                       for d in (train_data, val_data, test_data) if d)

        # In case the parent sort key isn't none
        if self.sort_key:
            for subset in splits:
                subset.sort_key = self.sort_key
        return splits

    def __getitem__(self, i):
        return self.examples[i]

    def __len__(self):
        try:
            return len(self.examples)
        except TypeError:
            return 2 ** 32

    def __iter__(self):
        for x in self.examples:
            yield x

    def __getattr__(self, attr):
        if attr in self.fields:
            for x in self.examples:
                yield getattr(x, attr)

    def filter_examples(self, field_names):
        """Remove unknown words from dataset examples with respect to given field.

        Arguments:
            field_names (list(str)): Within example only the parts with field names in
                field_names will have their unknown words deleted.
        """
        for i, example in enumerate(self.examples):
            for field_name in field_names:
                vocab = set(self.fields[field_name].vocab.stoi)
                text = getattr(example, field_name)
                example_part = [word for word in text if word in vocab]
                setattr(example, field_name, example_part)
            self.examples[i] = example


def check_split_ratio(split_ratio):
    """Check that the split ratio argument is not malformed"""
    valid_ratio = 0.
    if isinstance(split_ratio, float):
        # Only the train set relative ratio is provided
        # Assert in bounds, validation size is zero
        assert 0. < split_ratio < 1., (
            "Split ratio {} not between 0 and 1".format(split_ratio))

        test_ratio = 1. - split_ratio
        return (split_ratio, test_ratio, valid_ratio)
    elif isinstance(split_ratio, list):
        # A list of relative ratios is provided
        length = len(split_ratio)
        assert length == 2 or length == 3, (
            "Length of split ratio list should be 2 or 3, got {}".format(split_ratio))

        # Normalize if necessary
        ratio_sum = sum(split_ratio)
        if not ratio_sum == 1.:
            split_ratio = [float(ratio) / ratio_sum for ratio in split_ratio]

        if length == 2:
            return tuple(split_ratio + [valid_ratio])
        return tuple(split_ratio)
    else:
        raise ValueError('Split ratio must be float or a list, got {}'
                         .format(type(split_ratio)))


def stratify(examples, strata_field):
    # The field has to be hashable otherwise this doesn't work
    # There's two iterations over the whole dataset here, which can be
    # reduced to just one if a dedicated method for stratified splitting is used
    unique_strata = set(getattr(example, strata_field) for example in examples)
    strata_maps = {s: [] for s in unique_strata}
    for example in examples:
        strata_maps[getattr(example, strata_field)].append(example)
    return list(strata_maps.values())


def rationed_split(examples, train_ratio, test_ratio, val_ratio, rnd):
    """Create a random permutation of examples, then split them by ratios

    Arguments:
        examples: a list of data
        train_ratio, test_ratio, val_ratio: split fractions.
        rnd: a random shuffler

    Examples:
        >>> import ltp
        >>> examples = []
        >>> train_ratio, test_ratio, val_ratio = 0.7, 0.2, 0.1
        >>> rnd = ltp.data.dataset.RandomShuffler(None)
        >>> train_examples, test_examples, valid_examples = \
                ltp.data.dataset.rationed_split(examples, train_ratio,
                                                      test_ratio, val_ratio,
                                                      rnd)
    """
    N = len(examples)
    randperm = rnd(range(N))
    train_len = int(round(train_ratio * N))

    # Due to possible rounding problems
    if not val_ratio:
        test_len = N - train_len
    else:
        test_len = int(round(test_ratio * N))

    indices = (randperm[:train_len],  # Train
               randperm[train_len:train_len + test_len],  # Test
               randperm[train_len + test_len:])  # Validation

    # There's a possibly empty list for the validation set
    data = tuple([examples[i] for i in index] for index in indices)

    return data


class RandomShuffler(object):
    """Use random functions while keeping track of the random state to make it
    reproducible and deterministic."""

    def __init__(self, random_state=None):
        self._random_state = random_state
        if self._random_state is None:
            self._random_state = random.getstate()

    @contextmanager
    def use_internal_state(self):
        """Use a specific RNG state."""
        old_state = random.getstate()
        random.setstate(self._random_state)
        yield
        self._random_state = random.getstate()
        random.setstate(old_state)

    @property
    def random_state(self):
        return deepcopy(self._random_state)

    @random_state.setter
    def random_state(self, s):
        self._random_state = s

    def __call__(self, data):
        """Shuffle and return a new list."""
        with self.use_internal_state():
            return random.sample(data, len(data))



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>

class Error(Exception):
    pass


class RegistrationError(Error):
    pass


class DataUnsupported(Error):
    pass



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>

from typing import Tuple, Any
from itertools import chain

import torch
from . import Metric
from ltp.utils import length_to_mask
from ltp.utils.seqeval import get_entities


class BiaffineCRFSpan(Metric, alias='biaffine_crf_span'):
    """
    用于 语义角色标注
    """

    def __init__(self, id2label):
        super(BiaffineCRFSpan, self).__init__(f1=0., p=0., r=0.)
        self.field = id2label
        self.nb_correct = 0
        self.nb_pred = 0
        self.nb_true = 0

    def get_entities(self, labels):
        labels = labels.cpu().detach().numpy()
        labels = [self.field.vocab.itos[label] for label in labels]
        labels = get_entities(labels)

        return set(labels)

    def get_entities_with_list(self, labels):
        labels = [self.field.vocab.itos[label] for label in chain(*labels)]
        labels = get_entities(labels)
        return set(labels)

    def step(self, y_pred: Tuple[torch.Tensor, torch.Tensor, Any], y: Tuple[torch.Tensor, set]):
        rel_gold, rels_gold_set = y
        rels_scores, seq_lens, crf = y_pred

        mask = length_to_mask(seq_lens)
        mask = mask.unsqueeze_(-1).expand(-1, -1, mask.size(1))
        mask = mask & mask.transpose(-1, -2)
        mask = mask.flatten(end_dim=1)
        index = mask[:, 0]

        rel_gold = rel_gold.flatten(end_dim=1)[index]

        mask = mask[index]
        pred_entities = crf.decode(rels_scores.flatten(end_dim=1)[index], mask)

        rel_entities = self.get_entities(rel_gold[mask])
        pred_entities = self.get_entities_with_list(pred_entities)

        self.nb_correct += len(rel_entities & pred_entities)
        self.nb_pred += len(pred_entities)
        self.nb_true += len(rel_entities)

    @property
    def precision(self):
        return self.nb_correct / self.nb_pred if self.nb_pred > 0 else 0

    @property
    def recall(self):
        return self.nb_correct / self.nb_true if self.nb_true > 0 else 0

    @property
    def f1beta(self):
        p = self.precision
        r = self.recall
        score = 2 * p * r / (p + r) if (p + r > 0) else 0
        return score

    def compute(self):
        return {'f1': self.f1beta, 'p': self.precision, 'r': self.recall}

    def clear(self):
        self.nb_correct = 0
        self.nb_pred = 0
        self.nb_true = 0

    @classmethod
    def from_extra(cls, extra: dict):
        init = extra["config"]['init']
        id2label = init['id2label']
        for field_name, field in extra['fields']:
            if field_name == id2label:
                return {'id2label': field}

        return {'id2label': None}



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>
import torch.nn.functional as F
import torch
from typing import List
from . import Loss


class KdMseLoss(Loss, alias='kd_mse_loss'):
    def forward(self, logits_S, logits_T, temperature=1):
        '''
        Calculate the mse loss between logits_S and logits_T

        :param logits_S: Tensor of shape (batch_size, length, num_labels) or (batch_size, num_labels)
        :param logits_T: Tensor of shape (batch_size, length, num_labels) or (batch_size, num_labels)
        :param temperature: A float or a tensor of shape (batch_size, length) or (batch_size,)
        '''
        if isinstance(temperature, torch.Tensor) and temperature.dim() > 0:
            temperature = temperature.unsqueeze(-1)
        beta_logits_T = logits_T / temperature
        beta_logits_S = logits_S / temperature
        loss = F.mse_loss(beta_logits_S, beta_logits_T)
        return loss


class KdCeLoss(Loss, alias='kd_ce_loss'):
    def forward(self, logits_S, logits_T, temperature=1):
        '''
        Calculate the cross entropy between logits_S and logits_T

        :param logits_S: Tensor of shape (batch_size, length, num_labels) or (batch_size, num_labels)
        :param logits_T: Tensor of shape (batch_size, length, num_labels) or (batch_size, num_labels)
        :param temperature: A float or a tensor of shape (batch_size, length) or (batch_size,)
        '''
        if isinstance(temperature, torch.Tensor) and temperature.dim() > 0:
            temperature = temperature.unsqueeze(-1)
        beta_logits_T = logits_T / temperature
        beta_logits_S = logits_S / temperature
        p_T = F.softmax(beta_logits_T, dim=-1)
        loss = -(p_T * F.log_softmax(beta_logits_S, dim=-1)).sum(dim=-1).mean()
        return loss


class AttMseLoss(Loss, alias='att_mse_loss'):
    def forward(self, attention_S, attention_T, mask=None):
        '''
        Calculate the mse loss between attention_S and attention_T.

        :param logits_S: Tensor of shape  (batch_size, num_heads, length, length)
        :param logits_T: Tensor of shape  (batch_size, num_heads, length, length)
        :param mask:     Tensor of shape  (batch_size, length)
        '''
        if mask is None:
            attention_S_select = torch.where(attention_S <= -1e-3, torch.zeros_like(attention_S), attention_S)
            attention_T_select = torch.where(attention_T <= -1e-3, torch.zeros_like(attention_T), attention_T)
            loss = F.mse_loss(attention_S_select, attention_T_select)
        else:
            mask = mask.to(attention_S).unsqueeze(1).expand(-1, attention_S.size(1), -1)  # (bs, num_of_heads, len)
            valid_count = torch.pow(mask.sum(dim=2), 2).sum()
            loss = (F.mse_loss(attention_S, attention_T, reduction='none') * mask.unsqueeze(-1) * mask.unsqueeze(
                2)).sum() / valid_count
        return loss


class AttMseSumLoss(Loss, alias='att_mse_sum_loss'):
    def forward(self, attention_S, attention_T, mask=None):
        '''
        Calculate the mse loss between attention_S and attention_T, the dim of num_heads is summed

        :param logits_S: Tensor of shape  (batch_size, num_heads, length, length) or (batch_size, length, length)
        :param logits_T: Tensor of shape  (batch_size, num_heads, length, length) or (batch_size, length, length)
        :param mask:     Tensor of shape  (batch_size, length)
        '''
        if len(attention_S.size()) == 4:
            attention_T = attention_T.sum(dim=1)
            attention_S = attention_S.sum(dim=1)
        if mask is None:
            attention_S_select = torch.where(attention_S <= -1e-3, torch.zeros_like(attention_S), attention_S)
            attention_T_select = torch.where(attention_T <= -1e-3, torch.zeros_like(attention_T), attention_T)
            loss = F.mse_loss(attention_S_select, attention_T_select)
        else:
            mask = mask.to(attention_S)
            valid_count = torch.pow(mask.sum(dim=1), 2).sum()
            loss = (F.mse_loss(attention_S, attention_T, reduction='none') * mask.unsqueeze(-1) * mask.unsqueeze(
                1)).sum() / valid_count
        return loss


class AttCeLoss(Loss, alias='att_ce_loss'):
    def forward(self, attention_S, attention_T, mask=None):
        '''
        Calculate the cross entropy  between attention_S and attention_T.

        :param logits_S: Tensor of shape  (batch_size, num_heads, length, length)
        :param logits_T: Tensor of shape  (batch_size, num_heads, length, length)
        :param mask:     Tensor of shape  (batch_size, length)
        '''
        probs_T = F.softmax(attention_T, dim=-1)
        if mask is None:
            probs_T_select = torch.where(attention_T <= -1e-3, torch.zeros_like(attention_T), probs_T)
            loss = -((probs_T_select * F.log_softmax(attention_S, dim=-1)).sum(dim=-1)).mean()
        else:
            mask = mask.to(attention_S).unsqueeze(1).expand(-1, attention_S.size(1), -1)  # (bs, num_of_heads, len)
            loss = -((probs_T * F.log_softmax(attention_S, dim=-1) * mask.unsqueeze(2)).sum(
                dim=-1) * mask).sum() / mask.sum()
        return loss


class AttCeMeanLoss(Loss, alias='att_ce_mean_loss'):
    def forward(self, attention_S, attention_T, mask=None):
        '''
        Calculate the cross entropy  between attention_S and attention_T, the dim of num_heads is averaged

        :param logits_S: Tensor of shape  (batch_size, num_heads, length, length) or (batch_size, length, length)
        :param logits_T: Tensor of shape  (batch_size, num_heads, length, length) or (batch_size, length, length)
        :param mask:     Tensor of shape  (batch_size, length)
        '''
        if len(attention_S.size()) == 4:
            attention_S = attention_S.mean(dim=1)  # (bs, len, len)
            attention_T = attention_T.mean(dim=1)
        probs_T = F.softmax(attention_T, dim=-1)
        if mask is None:
            probs_T_select = torch.where(attention_T <= -1e-3, torch.zeros_like(attention_T), probs_T)
            loss = -((probs_T_select * F.log_softmax(attention_S, dim=-1)).sum(dim=-1)).mean()
        else:
            mask = mask.to(attention_S)
            loss = -((probs_T * F.log_softmax(attention_S, dim=-1) * mask.unsqueeze(1)).sum(
                dim=-1) * mask).sum() / mask.sum()
        return loss


class HidMseLoss(Loss, alias='hid_mse_loss'):
    def forward(self, state_S, state_T, mask=None):
        '''
        Calculate the mse loss between state_S and state_T, state is the hidden state of the model

        :param state_S: Tensor of shape  (batch_size, length, hidden_size)
        :param state_T: Tensor of shape  (batch_size, length, hidden_size)
        :param mask:    Tensor of shape  (batch_size, length)
        '''
        if mask is None:
            loss = F.mse_loss(state_S, state_T)
        else:
            mask = mask.to(state_S)
            valid_count = mask.sum() * state_S.size(-1)
            loss = (F.mse_loss(state_S, state_T, reduction='none') * mask.unsqueeze(-1)).sum() / valid_count
        return loss


class CosLoss(Loss, alias='cos_loss'):
    def forward(self, state_S, state_T, mask=None):
        '''
        This is the loss used in DistilBERT

        :param state_S: Tensor of shape  (batch_size, length, hidden_size)
        :param state_T: Tensor of shape  (batch_size, length, hidden_size)
        :param mask:    Tensor of shape  (batch_size, length)
        '''
        if mask is None:
            state_S = state_S.view(-1, state_S.size(-1))
            state_T = state_T.view(-1, state_T.size(-1))
        else:
            mask = mask.to(state_S).unsqueeze(-1).expand_as(state_S).to(torch.uint8)  # (bs,len,dim)
            state_S = torch.masked_select(state_S, mask).view(-1, mask.size(-1))  # (bs * select, dim)
            state_T = torch.masked_select(state_T, mask).view(-1, mask.size(-1))  # (bs * select, dim)

        target = state_S.new(state_S.size(0)).fill_(1)
        loss = F.cosine_embedding_loss(state_S, state_T, target, reduction='mean')
        return loss


class PkdLoss(Loss, alias='pkd_loss'):
    def forward(self, state_S, state_T, mask=None):
        '''
        This is the loss used in BERT-PKD

        :param state_S: Tensor of shape  (batch_size, length, hidden_size)
        :param state_T: Tensor of shape  (batch_size, length, hidden_size)
        '''

        cls_T = state_T[:, 0]  # (batch_size, hidden_dim)
        cls_S = state_S[:, 0]  # (batch_size, hidden_dim)
        normed_cls_T = cls_T / torch.norm(cls_T, dim=1, keepdim=True)
        normed_cls_S = cls_S / torch.norm(cls_S, dim=1, keepdim=True)
        loss = (normed_cls_S - normed_cls_T).pow(2).sum(dim=-1).mean()
        return loss


class FspLoss(Loss, alias='fsp_loss'):
    def forward(state_S: List[torch.Tensor], state_T: List[torch.Tensor], mask=None):
        '''
        Variant of FSP loss from A Gift from Knowledge Distillation: Fast Optimization, Network Minimization and Transfer Learning

        :param state_S: list of two tensors, each tensor is of shape  (batch_size, length, hidden_size)
        :param state_T: list of two tensors, each tensor is of shape  (batch_size, length, hidden_size)
        :param mask:    Tensor of shape  (batch_size, length)
        '''
        if mask is None:
            state_S_0 = state_S[0]  # (batch_size , length, hidden_dim)
            state_S_1 = state_S[1]  # (batch_size,  length, hidden_dim)
            state_T_0 = state_T[0]
            state_T_1 = state_T[1]
            gram_S = torch.bmm(state_S_0.transpose(1, 2), state_S_1) / state_S_1.size(
                1)  # (batch_size, hidden_dim, hidden_dim)
            gram_T = torch.bmm(state_T_0.transpose(1, 2), state_T_1) / state_T_1.size(1)
        else:
            mask = mask.to(state_S[0]).unsqueeze(-1)
            lengths = mask.sum(dim=1, keepdim=True)
            state_S_0 = state_S[0] * mask
            state_S_1 = state_S[1] * mask
            state_T_0 = state_T[0] * mask
            state_T_1 = state_T[1] * mask
            gram_S = torch.bmm(state_S_0.transpose(1, 2), state_S_1) / lengths
            gram_T = torch.bmm(state_T_0.transpose(1, 2), state_T_1) / lengths
        loss = F.mse_loss(gram_S, gram_T)
        return loss


class MmdLoss(Loss, alias='mmd_loss'):
    def forward(state_S: List[torch.Tensor], state_T: List[torch.Tensor], mask=None):
        '''
        Variant of  NST loss and its from Like What You Like: Knowledge Distill via Neuron Selectivity Transfer

        :param state_S: list of two tensors, each tensor is of shape  (batch_size, length, hidden_size)
        :param state_T: list of two tensors, each tensor is of shape  (batch_size, length, hidden_size)
        :param mask:    Tensor of shape  (batch_size, length)
        '''

        state_S_0 = state_S[0]  # (batch_size , length, hidden_dim_S)
        state_S_1 = state_S[1]  # (batch_size , length, hidden_dim_S)
        state_T_0 = state_T[0]  # (batch_size , length, hidden_dim_T)
        state_T_1 = state_T[1]  # (batch_size , length, hidden_dim_T)
        if mask is None:
            gram_S = torch.bmm(state_S_0, state_S_1.transpose(1, 2)) / state_S_1.size(2)  # (batch_size, length, length)
            gram_T = torch.bmm(state_T_0, state_T_1.transpose(1, 2)) / state_T_1.size(2)
            loss = F.mse_loss(gram_S, gram_T)
        else:
            mask = mask.to(state_S[0])
            valid_count = torch.pow(mask.sum(dim=1), 2).sum()
            gram_S = torch.bmm(state_S_0, state_S_1.transpose(1, 2)) / state_S_1.size(1)  # (batch_size, length, length)
            gram_T = torch.bmm(state_T_0, state_T_1.transpose(1, 2)) / state_T_1.size(1)
            loss = (F.mse_loss(gram_S, gram_T, reduction='none') * mask.unsqueeze(-1) * mask.unsqueeze(
                1)).sum() / valid_count
        return loss



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>
from torch import nn

from ltp.core import Registrable


class Loss(nn.modules.loss._Loss, metaclass=Registrable):
    def __init__(self, *args, **kwargs):
        super(Loss, self).__init__(*args, **kwargs)

    def forward(self, *input, **kwargs):
        raise NotImplementedError()

    def distill(self, inputs, targets, temperature_calc, distill_loss, gold=None):
        raise NotImplementedError()


Loss.weak_register("L1Loss", nn.L1Loss)
Loss.weak_register("L2Loss", nn.MSELoss)
Loss.weak_register("CTCLoss", nn.CTCLoss)
Loss.weak_register("NLLLoss", nn.NLLLoss)
Loss.weak_register("BCELoss", nn.BCELoss)
Loss.weak_register("MSELoss", nn.MSELoss)
Loss.weak_register("KLDivLoss", nn.KLDivLoss)
Loss.weak_register("NLLLoss2d", nn.NLLLoss2d)
Loss.weak_register("SmoothL1Loss", nn.SmoothL1Loss)
Loss.weak_register("PoissonNLLLoss", nn.PoissonNLLLoss)
Loss.weak_register("SoftMarginLoss", nn.SoftMarginLoss)
Loss.weak_register("CrossEntropyLoss", nn.CrossEntropyLoss)
Loss.weak_register("MarginRankingLoss", nn.MarginRankingLoss)
Loss.weak_register("BCEWithLogitsLoss", nn.BCEWithLogitsLoss)
Loss.weak_register("TripletMarginLoss", nn.TripletMarginLoss)
Loss.weak_register("HingeEmbeddingLoss", nn.HingeEmbeddingLoss)
Loss.weak_register("CosineEmbeddingLoss", nn.CosineEmbeddingLoss)
Loss.weak_register("MultiLabelMarginLoss", nn.MultiLabelMarginLoss)

from . import kd_loss
from . import task_loss

__all__ = ['Loss']
